import abc
from typing import Union, Optional, List
import copy
from .exceptions import BadStatement

specialChars = ['\\', '.', '^', '$', '|', '?', '*', '+', '(', ')', '[', ']', '{', '}', '-']
specialCharsd = {i:'\\'+i for i in specialChars}

def scape_special(values):
    nvalues = ""
    for i in values:
        if i in specialCharsd:
            nvalues += specialCharsd[i]
        else:
            nvalues += i
    return nvalues

### Base
class Base(metaclass=abc.ABCMeta):
    __slosts__ = ('_requiresGroup')
    def __init__(self, requiresGroup=False):
        self._requiresGroup = requiresGroup

    @property
    def isOneCharLiteral(self):
        return False

    @property
    def isLiteral(self):
        return False

    @property
    def isNegatedLiteral(self):
        return False

    @property
    def requiresGroup(self):
        return self._requiresGroup

    @abc.abstractproperty
    def value(self): # pragma: no cover
        raise NotImplementedError

    def copy(self): # pragma: no cover
        return copy.deepcopy(self)

class Literal(Base, metaclass=abc.ABCMeta):
    @property
    def isLiteral(self):
        return True

class NegatedLiteral(Literal, metaclass=abc.ABCMeta):
    __slosts__ = ()
    @property
    def isNegatedLiteral(self):
        return True

class OneCharLiteral(Literal, metaclass=abc.ABCMeta):
    __slosts__ = ()
    @property
    def isOneCharLiteral(self):
        return True

class BaseGroups(Base, metaclass=abc.ABCMeta):
    __slosts__ = ('_elements')
    def __init__(self, elements):
        super().__init__()
        self._elements = elements

    @classmethod
    def element(cls, elements): # pragma: no cover
        return cls(elements)


class BaseQuantifier(BaseGroups, metaclass=abc.ABCMeta):
    __slosts__ = ()
    def __init__(self, element):
        assert not isinstance(element, (list, tuple))
        super().__init__(element)

    def _value(self):
        inner = self._elements.value
        verif = not(len(inner) == 1 or (len(inner) == 2 and '\\' in inner))
        grouping = f"(?:{inner})" if self._elements.requiresGroup and verif else inner
        return grouping

### Grouping
class Root(BaseGroups):
    __slosts__ = ()
    @property
    def value(self):
        return "".join((i.value for i in self._elements))

class SetOfLiterals(BaseGroups):
    __slosts__ = ()
    def __init__(self, elements):
        super().__init__(elements)
        # the next line work but is imposible report properly the error make by the user
        # assert all(map(lambda x: x.isLiteral and not x.isNegatedLiteral ,elements))
        for x in elements:
            if not x.isLiteral or x.isNegatedLiteral:
                raise BadStatement(f'type {type(x)} isn\'t admited inside "setOfLiterals"')

    @property
    def value(self):
        res = "["
        for i in self._elements:
            value = i.value
            if not i.isOneCharLiteral:
                value = value[1:-1]
            res += value
        res += "]"
        return res

class Capture(BaseGroups):
    __slosts__ = ()
    @property
    def value(self):
        return "("+"".join((i.value for i in self._elements))+")"

class Group(BaseGroups):
    __slosts__ = ()
    @property
    def value(self):
        return "(?:"+"".join((i.value for i in self._elements))+")"

### Quatifier
class OptionalQ(BaseQuantifier):
    __slosts__ = ()
    @property
    def value(self):
        return self._value() + '?'

class zeroOrMore(BaseQuantifier):
    __slosts__ = ()
    @property
    def value(self):
        return self._value() + '*'

class zeroOrMoreLazy(BaseQuantifier):
    __slosts__ = ()
    @property
    def value(self):
        return self._value() + '*?'

class oneOrMore(BaseQuantifier):
    __slosts__ = ()
    @property
    def value(self):
        return self._value() + '+'

class oneOrMoreLazy(BaseQuantifier):
    __slosts__ = ()
    @property
    def value(self):
        return self._value() + '+?'

class Between(BaseQuantifier):
    __slosts__ = ('_a', '_b')
    def __init__(self, element, a: int, b: int):
        assert isinstance(a, int) and isinstance(b, int), "Between params must be int"
        assert 0 <= a <= b
        super().__init__(element)
        self._a = str(a)
        self._b = str(b)

    @property
    def value(self):
        if self._a == '0':
            if self._b == '1':
                return self._value() + '?'
            return self._value() + '{,' + self._b + '}'
        return self._value() + '{' + self._a + ',' + self._b + '}'

class betweenLazy(BaseQuantifier):
    __slosts__ = ('_a', '_b')
    def __init__(self, element, a: int, b: int):
        assert isinstance(a, int) and isinstance(b, int), "betweenLazy params must be int"
        assert 0 <= a <= b
        super().__init__(element)
        self._a = str(a)
        self._b = str(b)

    @property
    def value(self):
        return self._value() + '{' + self._a + ',' + self._b + '}?'

class Exactly(BaseQuantifier):
    __slosts__ = ('_a')
    def __init__(self, element, a: int):
        assert isinstance(a, int), "Exactly param must be int"
        assert a > 0
        super().__init__(element)
        self._a = str(a)

    @property
    def value(self):
        return self._value() + '{' + self._a + '}'

class atLeast(BaseQuantifier):
    __slosts__ = ('_a')
    def __init__(self, element, a: int):
        assert isinstance(a, int), "Exactly param must be int"
        assert a >= 0
        super().__init__(element)
        self._a = str(a)

    @property
    def value(self):
        if self._a == '0':
            return self._value() + '*'
        if self._a == '1':
            return self._value() + '+'
        return self._value() + '{' + self._a + ',}'

class upTo(BaseQuantifier):
    __slosts__ = ('_b')
    def __init__(self, element, b: int):
        assert isinstance(b, int), "Exactly param must be int"
        assert b > 0
        super().__init__(element)
        self._b = str(b)

    @property
    def value(self):
        if self._b == '1':
            return self._value() + '?'
        return self._value() + '{,' + self._b + '}'

 ### Literals
class anyChar(Base):
    __slosts__ = ()
    @property
    def value(self):
        return '.'

class whitespaceChar(OneCharLiteral):
    __slosts__ = ()
    @property
    def value(self):
        return '\\s'

class nonWhitespaceChar(OneCharLiteral):
    __slosts__ = ()
    @property
    def value(self):
        return '\\S'

class Digit(OneCharLiteral):
    __slosts__ = ()
    @property
    def value(self):
        return '\\d'

class nonDigit(OneCharLiteral):
    __slosts__ = ()
    @property
    def value(self):
        return '\\D'

class Word(OneCharLiteral):
    __slosts__ = ()
    @property
    def value(self):
        return '\\w'

class nonWord(OneCharLiteral):
    __slosts__ = ()
    @property
    def value(self):
        return '\\W'

class Char(OneCharLiteral):
    __slosts__ = tuple(['_value'])
    def __init__(self, value):
        assert isinstance(value, str) and len(value) == 1
        super().__init__()
        self._value = scape_special(value)

    @property
    def value(self):
        return self._value

class rawChar(Base):
    __slosts__ = tuple(['_value'])
    def __init__(self, value):
        assert isinstance(value, str) and len(value) == 1
        super().__init__()
        self._value = value

    @property
    def value(self):
        return self._value

class String(Base):
    __slosts__ = tuple(['_values'])
    def __init__(self, values):
        assert isinstance(values, str)
        super().__init__(requiresGroup=True)
        self._values = scape_special(values)

    @property
    def value(self):
        return self._values

class rawString(Base):
    __slosts__ = tuple(['_values'])
    def __init__(self, values):
        assert isinstance(values, str)
        super().__init__(requiresGroup=True)
        self._values = values

    @property
    def value(self):
        return self._values

class Newline(OneCharLiteral):
    __slosts__ = ()

    @property
    def value(self):
        return '\\n'

class carriageReturn(OneCharLiteral):
    __slosts__ = ()

    @property
    def value(self):
        return '\\r'

class Tab(OneCharLiteral):
    __slosts__ = ()

    @property
    def value(self):
        return '\\t'

class Space(OneCharLiteral):
    __slosts__ = ()

    @property
    def value(self):
        return ' '

class Range(Literal):
    __slosts__ = ('_begin', '_end', '_exclude', '_expression')
    def __init__(self, begin: Union[str, int], end: Union[str, int],
                 exclude: Optional[Union[List[Union[str, int]], str]]=None):
        super().__init__()
        tb = type(begin)
        # TODO: write explanations of asserts
        assert (isinstance(begin, str) and len(begin) == 1 and begin.isalnum()) or\
               (isinstance(begin, int) and 0 <= begin <= 9)
        assert (isinstance(end, str) and len(end) == 1 and begin.isalnum()) or\
               (isinstance(end, int) and 0 <= end <= 9)
        assert begin <= end
        assert str(begin).islower() == str(end).islower()
        assert str(begin).isnumeric() == str(end).isnumeric()
        if exclude:
            if isinstance(exclude, str):
                exclude = list(exclude)
            assert all([((isinstance(i, str) and len(i) == 1 and begin.isalnum()) or\
                   (isinstance(i, int) and 0 <= i <= 9)) and isinstance(i, tb) for i in exclude])
            exclude = list(sorted(exclude))
        self._begin = begin
        self._end = end
        self._exclude = exclude
        self._expression = ""
        if isinstance(begin, int):
            if exclude:
                self._expression = '['
                excludec = set(exclude)
                for i in range(begin, end+1):
                    if i in excludec:
                        continue
                    self._expression += str(i)
                self._expression += ']'
            else:
                self._expression = '['+str(begin)+'-'+str(end)+']'
        else:
            if exclude:
                excludec = set(exclude)
                self._expression = '['
                for i in range(ord(begin), ord(end)+1):
                    ch = chr(i)
                    if ch in excludec:
                        continue
                    self._expression += ch
                self._expression += ']'
            else:
                self._expression = '['+begin+'-'+end+']'

    @property
    def value(self):
        return self._expression

class anythingButRange(NegatedLiteral):
    __slosts__ = ('_begin', '_end', '_exclude', '_expression')
    def __init__(self, begin: Union[str, int], end: Union[str, int],
                 exclude: Optional[Union[List[Union[str, int]], str]]=None):
        super().__init__()
        tb = type(begin)
        # TODO: write explanations of asserts
        assert (isinstance(begin, str) and len(begin) == 1 and begin.isalnum()) or\
               (isinstance(begin, int) and 0 <= begin <= 9)
        assert (isinstance(end, str) and len(end) == 1 and begin.isalnum()) or\
               (isinstance(end, int) and 0 <= end <= 9)
        assert begin <= end
        assert str(begin).islower() == str(end).islower()
        assert str(begin).isnumeric() == str(end).isnumeric()
        if exclude:
            if isinstance(exclude, str):
                exclude = list(exclude)
            assert all([((isinstance(i, str) and len(i) == 1 and begin.isalnum()) or\
                   (isinstance(i, int) and 0 <= i <= 9)) and isinstance(i, tb) for i in exclude])
            exclude = list(sorted(exclude))
        self._begin = begin
        self._end = end
        self._exclude = exclude
        self._expression = ""
        if isinstance(begin, int):
            if exclude:
                self._expression = '[^'
                excludec = set(exclude)
                for i in range(begin, end+1):
                    if i in excludec:
                        continue
                    self._expression += str(i)
                self._expression += ']'
            else:
                self._expression = '[^'+str(begin)+'-'+str(end)+']'
        else:
            if exclude:
                excludec = set(exclude)
                self._expression = '[^'
                for i in range(ord(begin), ord(end)+1):
                    ch = chr(i)
                    if ch in excludec:
                        continue
                    self._expression += ch
                self._expression += ']'
            else:
                self._expression = '[^'+begin+'-'+end+']'

    @property
    def value(self):
        return self._expression


class anyOfChars(Literal):
    __slosts__ = ('_values')
    def __init__(self, values):
        assert isinstance(values, str)
        super().__init__()
        self._values = scape_special(values)

    @property
    def value(self):
        return '['+self._values+']'

class anythingButChars(NegatedLiteral):
    __slosts__ = ('_values')
    def __init__(self, values):
        assert isinstance(values, str)
        super().__init__()
        self._values = scape_special(values)

    @property
    def value(self):
        return '[^'+self._values+']'



import re
import copy
from typing import Union, Optional, List
from .exceptions import BadStatement
from .aux_classes import Stack, StackFrame
### Grouping
from .regex_classes import SetOfLiterals
from .regex_classes import Capture, Group
### Quatifiers
from .regex_classes import OptionalQ
from .regex_classes import zeroOrMore, zeroOrMoreLazy
from .regex_classes import oneOrMore, oneOrMoreLazy
from .regex_classes import Between, betweenLazy
from .regex_classes import atLeast, upTo
from .regex_classes import Exactly
### Others
from .regex_classes import Root
### Literals
from .regex_classes import anyChar, whitespaceChar
from .regex_classes import nonWhitespaceChar, Digit
from .regex_classes import nonDigit, Word, nonWord
from .regex_classes import Char, rawChar, String, rawString
from .regex_classes import Newline, carriageReturn, Tab, Space
from .regex_classes import Range, anythingButRange
from .regex_classes import anyOfChars, anythingButChars



class ExpressiveRegex:
    __slosts__ = ('_expression', '_hasDefineStart', '_hasDefineEnd', '_flags', '_stack')
    def __init__(self, mutable=False):
        self._expression = ""
        self._hasDefineStart = False
        self._hasDefineEnd = False
        self._flags = {
            # TODO: add the other flags
            'i': False, # case-insensitive matching
            'm': False, # multiline
        }
        self._stack = Stack()
        self._stack.push(StackFrame(Root))
        self._mutable = mutable

    @property
    def mutable(self):
        return self._mutable

    @property
    def _currentFrame(self): # pragma: no cover
        return self._stack.top()

    def _applyQuatifier(self, element): # pragma: no cover
        if self._currentFrame.quantifier:
            wrap = self._currentFrame.get_qinstance(element)
            self._currentFrame.quantifier = None
            self._currentFrame._quantifier_args = ()
            self._currentFrame._quantifier_kwargs = {}
        else:
            wrap = element
        return wrap

    def _matchElement(self, cls, *args, **kwargs): # pragma: no cover
        element = cls(*args, **kwargs)
        wrap = self._applyQuatifier(element)
        self._currentFrame.elements.append(wrap)
        return self

    def _instance(self): # pragma: no cover
        if self._mutable:
            return self
        obj = copy.deepcopy(self)
        obj._expression = ""
        return obj

    def _quantifier(self, qclass, *args, **kwargs):
        if self._currentFrame.quantifier is not None:
            raise BadStatement(f'cannot quantify regular expression with "{str(qclass)}" because it\'s already being quantified with "{self._currentFrame.quantifier}"')
        if self._currentFrame.type is SetOfLiterals:
            raise BadStatement('Quatifiers aren\'t admited inside "setOfLiterals".')
        instance = self._instance()
        instance._currentFrame.quantifier = qclass
        instance._currentFrame._quantifier_args = args
        instance._currentFrame._quantifier_kwargs = kwargs
        return instance

    def _group(self, gclass):
        instance = self._instance()
        newFrame = StackFrame(gclass)
        instance._stack.push(newFrame)
        return instance

    ### Quantifiers

    @property
    def optional(self):
        return self._quantifier(OptionalQ)

    @property
    def zeroOrMore(self):
        return self._quantifier(zeroOrMore)

    @property
    def zeroOrMoreLazy(self):
        return self._quantifier(zeroOrMoreLazy)

    @property
    def oneOrMore(self):
        return self._quantifier(oneOrMore)

    @property
    def oneOrMoreLazy(self):
        return self._quantifier(oneOrMoreLazy)

    def exactly(self, a: int):
        return self._quantifier(Exactly, a)

    def between(self, a: int, b: int):
        return self._quantifier(Between, a, b)

    def betweenLazy(self, a: int, b: int):
        return self._quantifier(betweenLazy, a, b)

    def atLeast(self, a: int):
        return self._quantifier(atLeast, a)

    def upTo(self, b: int):
        return self._quantifier(upTo, b)

    ### Grouping

    def end(self):
        if len(self._stack) == 1:
            raise BadStatement("Cannot call end while building the root expression.")
        instance = self._instance()
        frame = instance._stack.pop()
        wrap = instance._applyQuatifier(frame.get_instance())
        instance._currentFrame.append(wrap)
        return instance

    @property
    def setOfLiterals(self):
        return self._group(SetOfLiterals)

    @property
    def capture(self):
        return self._group(Capture)

    @property
    def group(self):
        return self._group(Group)

    ### Literals

    @property
    def anyChar(self):
        instance = self._instance()
        return instance._matchElement(anyChar)

    @property
    def whitespaceChar(self):
        instance = self._instance()
        return instance._matchElement(whitespaceChar)

    @property
    def nonWhitespaceChar(self):
        instance = self._instance()
        return instance._matchElement(nonWhitespaceChar)

    @property
    def digit(self):
        instance = self._instance()
        return instance._matchElement(Digit)

    @property
    def nonDigit(self):
        instance = self._instance()
        return instance._matchElement(nonDigit)

    @property
    def word(self):
        instance = self._instance()
        return instance._matchElement(Word)

    @property
    def nonWord(self):
        instance = self._instance()
        return instance._matchElement(nonWord)

    @property
    def newline(self):
        instance = self._instance()
        return instance._matchElement(Newline)

    @property
    def carriageReturn(self):
        instance = self._instance()
        return instance._matchElement(carriageReturn)

    @property
    def tab(self):
        instance = self._instance()
        return instance._matchElement(Tab)

    @property
    def space(self):
        instance = self._instance()
        return instance._matchElement(Space)


    def char(self, value):
        instance = self._instance()
        return instance._matchElement(Char, value)

    def rawChar(self, value):
        instance = self._instance()
        return instance._matchElement(rawChar, value)

    def string(self, value):
        instance = self._instance()
        return instance._matchElement(String, value)

    def rawString(self, value):
        instance = self._instance()
        return instance._matchElement(rawString, value)

    def range(self, begin: Union[str, int], end: Union[str, int],
              exclude: Optional[Union[List[Union[str, int]], str]]=None):
        instance = self._instance()
        return instance._matchElement(Range, begin, end, exclude)

    def anythingButRange(self, begin: Union[str, int], end: Union[str, int],
                         exclude: Optional[Union[List[Union[str, int]], str]]=None):
        instance = self._instance()
        return instance._matchElement(anythingButRange, begin, end, exclude)

    def anyOfChars(self, value):
        instance = self._instance()
        return instance._matchElement(anyOfChars, value)

    def anythingButChars(self, value):
        instance = self._instance()
        return instance._matchElement(anythingButChars, value)

    ### Build Expression

    def toRegexString(self)->String:
        txt = 'Cannot compute the value of a not yet fully specified regex object.\n'
        txt += f'(Try adding a .end() call to match the "{str(self._currentFrame.type)}")\n'
        assert len(self._stack) == 1, txt
        if not self._mutable and self._expression:
            return self._expression
        exp = self._currentFrame.get_instance().value
        self._expression = exp
        return exp

    def toRegex(self):
        exp = self.toRegexString()
        return re.compile(exp)



import copy

class Stack: # pragma: no cover
    __slots__ = tuple(['_stack'])

    def __init__(self):
        self._stack = []

    def copy(self):
        res = Stack()
        res._stack = copy.deepcopy(self._stack)
        return res

    def __len__(self):
        return len(self._stack)

    @property
    def empty(self):
        return len(self._stack) == 0

    def isempty(self):
        return len(self._stack) == 0

    def top(self):
        if len(self._stack) == 0:
            raise IndexError('pop from empty Stack')
        return self._stack[-1]

    def push(self, value):
        self._stack.append(value)

    def pop(self):
        if len(self._stack) == 0:
            raise IndexError('pop from empty Stack')
        val = self._stack.pop()
        return val


class StackFrame: # pragma: no cover
    __slots__ = ('_elements', '_quantifier', '_type',
                 '_quantifier_args', '_quantifier_kwargs')
    def __init__(self, cls):
        self._elements = []
        self._quantifier = None
        self._quantifier_args = ()
        self._quantifier_kwargs = {}
        self._type = cls

    def get_instance(self):
        return self._type(self._elements)

    def get_qinstance(self, instance=None):
        args = self._quantifier_args
        kwargs = self._quantifier_kwargs
        if instance:
            return self._quantifier(instance, *args, **kwargs)
        return self._quantifier(self.get_instance(), *args, **kwargs)

    def append(self, element):
        self._elements.append(element)

    @property
    def elements(self):
        return self._elements

    @property
    def quantifier(self):
        return self._quantifier

    @property
    def type(self):
        return self._type

    @quantifier.setter
    def quantifier(self, value):
        self._quantifier = value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['expressive_regex']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'expressive-regex',
    'version': '0.4.5',
    'description': '',
    'long_description': '# Expressive Regex\n\n<img alt="PyPI - License" src="https://img.shields.io/github/license/fsadannn/expressive_regex"> <img alt="PyPI - License" src="https://travis-ci.org/fsadannn/expressive_regex.svg"> <img alt="Codecov" src="https://img.shields.io/codecov/c/github/fsadannn/expressive_regex.svg">\n\nThis project was made with inspiration from [Super Expressive for JavaScript](https://github.com/francisrstokes/super-expressive).\n\nExpressive Regex allow you to build regular expressions in almost natural language and without external dependency.\n\n[Documentation](https://fsadannn.github.io/expressive_regex/)\n\n**Example**\nto match a telephone number that can be in the format 555-555-555, 555 555 555 or 555555555.\n\n```Python\nExpressiveRegex()\\\n    .exactly(2).group\\\n        .oneOrMore.digit\\\n        .optional.setOfLiterals\\\n            .char(\'-\')\\\n            .whitespaceChar\\\n        .end()\\\n    .end()\\\n    .oneOrMore.digit\\\n.toRegexString()\n```\n\n```Python\n"(?:\\d+[\\-\\s]?){2}\\d+"\n```',
    'author': 'fsadannn',
    'author_email': 'fsadannn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.1,<4',
}


setup(**setup_kwargs)
